<?php

/**
 * @file
 * Database interface code for automatic slave selection depending on type of
 * query.
 *
 * @todo Ensure default method arguments will work if they differ from what's
 *       defined in class Database.
 */

/**
 * Default replication lag assumed (2 seconds).
 */
define('AUTOSLAVE_ASSUMED_REPLICATION_LAG', 2);

/**
 * @ingroup database
 * @{
 */

/**
 * Specific auto slave implementation of DatabaseConnection.
 */
class DatabaseConnection_autoslave extends DatabaseConnection {

  /**
   * List of tables that should use "master" as target
   */
  protected $__tables;

  /**
   * Targets for master and slave
   */
  protected $__master;
  protected $__slave;

  /**
   * Force queries to master
   */
  private $__force_master = 0;


  /**
   * Constructor.
   */
  public function __construct(array $connection_options = array()) {
    $this->__master = !empty($connection_options['master']) ? $connection_options['master'] : 'master';
    $this->__slave = !empty($connection_options['slave']) ? $connection_options['slave'] : 'autoslave';
    $this->__tables = !empty($connection_options['tables']) ? $connection_options['tables'] : array('sessions', 'sempahore');

    // Initialize force master tables
    if (!empty($this->__tables)) {
      $this->__tables = array_combine($this->__tables, $this->__tables);
    }

    // Has master been forced before Database bootstrapping?
    if (!empty($_GLOBALS['autoslave_pre_bootstrap_force_master'])) {
      $this->forceMaster(1);
    }
  }

  /**
   * Dispatch all methods defined in class PDO to the appropiate backend
   */
  public function beginTransaction() { return Database::getConnection($this->__master)->beginTransaction(); }
  public function errorCode() { return Database::getConnection($this->__master)->errorCode(); }
  public function errorInfo() { return Database::getConnection($this->__master)->errorInfo(); }
  public function exec($statement) { return Database::getConnection($this->__master)->exec($statement); }
  public function getAttribute($attribute) { return Database::getConnection($this->__master)->getAttribute($attribute); }
  static public function getAvailableDrivers() { $drivers = Database::getConnection($this->__master)->getAvailableDrivers(); $drivers[] = 'autoslave'; return $drivers; }
  public function lastInsertId($name = NULL) { return Database::getConnection($this->__master)->lastInsertId($name); }
  public function quote() { return Database::getConnection($this->__master)->quote(); }
  public function setAttribute($attribute, $value) { return Database::getConnection($this->__master)->setAttribute($attribute, $value); }

  /**
   * Dispatch all methods defined in class Database to the appropiate backend
   */
  /**
   * "master" functions
   */
  public function inTransaction() { return Database::getConnection($this->__master)->inTransaction(); }
  public function transactionDepth() { return Database::getConnection($this->__master)->transactionDepth(); }
  public function rollback($savepoint_name = 'drupal_transaction') { return Database::getConnection($this->__master)->rollback($savepoint_name); }
  public function pushTransaction($name) { return Database::getConnection($this->__master)->pushTransaction($name); }
  public function popTransaction($name) { return Database::getConnection($this->__master)->popTransaction($name); }
  protected function popCommittableTransactions() { return Database::getConnection($this->__master)->popCommittableTransactions(); }
  protected function generateTemporaryTableName() { return Database::getConnection($this->__master)->generateTemporaryTableName(); }
  public function supportsTransactions() { return Database::getConnection($this->__master)->supportsTransactions(); }
  public function supportsTransactionalDDL() { return Database::getConnection($this->__master)->supportsTransactions(); }
  function commit() { return Database::getConnection($this->__master)->commit(); }

  /**
   * "slave" functions
   */
  protected function defaultOptions() { return Database::getConnection($this->__slave)->defaultOptions(); }
  public function getConnectionOptions() { return Database::getConnection($this->__slave)->getConnectionOptions(); }
  protected function setPrefix($prefix) { return Database::getConnection($this->__slave)->setPrefix($prefix); }
  public function prefixTables($sql) { return Database::getConnection($this->__slave)->prefixTables($sql); }
  public function tablePrefix($table = 'default') { return Database::getConnection($this->__slave)->tablePrefix($table); }
  public function prepareQuery($query) { return Database::getConnection($this->__slave)->prepareQuery($query); }
  public function setLogger(DatabaseLog $logger) { return Database::getConnection($this->__slave)->setLogger($logger); }
  public function getLogger() { return Database::getConnection($this->__slave)->getLogger(); }
  public function makeSequenceName($table, $field) { return Database::getConnection($this->__slave)->makeSequenceName($table, $field); }
  public function makeComment($comments) { return Database::getConnection($this->__slave)->makeComment($comments); }
  protected function filterComment($comment = '') { return Database::getConnection($this->__slave)->filterComment($comment); }
  protected function expandArguments(&$query, &$args) { return Database::getConnection($this->__slave)->expandArguments($query, $args); }
  public function escapeTable($table) { return Database::getConnection($this->__slave)->escapeTable($table); }
  public function escapeField($field) { return Database::getConnection($this->__slave)->escapeField($field); }
  public function escapeAlias($field) { return Database::getConnection($this->__slave)->escapeAlias($field); }
  public function escapeLike($string) { return Database::getConnection($this->__slave)->escapeLike($string); }

  /**
   * The following methods are absolutely necessary to overload manually (at least for MySQL)
   */

  #public function getDriverClass($class, array $files = array(), $use_autoload = FALSE) { return Database::getConnection($this->__master)->getDriverClass($class, $files, $use_autoload); }
  public function version() { return Database::getConnection($this->__slave)->version(); }
  public function schema() { return Database::getConnection($this->__master)->schema(); }
  public function driver() { return 'autoslave'; }
  public function databaseType() { return Database::getConnection($this->__slave)->databaseType(); }
  public function query($query, array $args = array(), $options = array()) { 
    if (!$this->prepareAutoslaveTarget($options)) {
      $options['target'] = $this->deriveTargetFromQuery($query);
    }
    return Database::getConnection($options['target'])->query($query, $args, $options);
  }
  public function queryRange($query, $from, $count, array $args = array(), array $options = array()) { 
    if (!$this->prepareAutoslaveTarget($options)) {
      $options['target'] = $this->deriveTargetFromQuery($query);
    }
    return Database::getConnection($options['target'])->queryRange($query, $from, $count, $args, $options);
  }
  public function queryTemporary($query, array $args = array(), array $options = array()) {
    if (!$this->prepareAutoslaveTarget($options)) {
      $options['target'] = $this->__master;
    }
    return Database::getConnection($options['target'])->queryTemporary($query, $args, $options);
  }
  public function mapConditionOperator($operator) { return Database::getConnection($this->__master)->mapConditionOperator($operator); }
  public function nextId($existing_id = 0) { return Database::getConnection($this->__master)->nextId($existing_id); }

  public function select($table, $alias = NULL, array $options = array()) {
    if (!$this->prepareAutoslaveTarget($options)) {
      $options['target'] = $this->getTargetForTable($table);
    }
    return Database::getConnection($options['target'])->select($table, $alias, $options);
  }
  public function insert($table, $options = array()) {
    if (!$this->prepareAutoslaveTarget($options)) {
      $options['target'] = $this->__master;
      $this->addAffectedTable($table);
    }
    return Database::getConnection($options['target'])->insert($table, $options);
  }

  public function merge($table, $options = array()) {
    if (!$this->prepareAutoslaveTarget($options)) {
      $options['target'] = $this->__master;
      $this->addAffectedTable($table);
    }
    return Database::getConnection($options['target'])->merge($table, $options);
  }

  public function update($table, $options = array()) {
    if (!$this->prepareAutoslaveTarget($options)) {
      $options['target'] = $this->__master;
      $this->addAffectedTable($table);
    }
    return Database::getConnection($options['target'])->update($table, $options);
  }

  public function delete($table, $options = array()) {
    if (!$this->prepareAutoslaveTarget($options)) {
      $options['target'] = $this->__master;
      $this->addAffectedTable($table);
    }
    return Database::getConnection($options['target'])->delete($table, $options);
  }

  public function truncate($table, $options = array()) {
    if (!$this->prepareAutoslaveTarget($options)) {
      $options['target'] = $this->__master;
      $this->addAffectedTable($table);
    }
    return Database::getConnection($options['target'])->truncate($table, $options);
  }

  public function startTransaction($name = '') {
    return new DatabaseTransaction_autoslave($this, $name);
  }


  /**
   * Magic methods. Supports propeties/functions/methods not defined by the Database class
   */
  public static function __callStatic($method, $args) {
    return call_user_func_array(array(Database::getConnection($this->__master)->get_class_name(), $method), $args);
  }

  public function __call($method, $args) {
    return call_user_func_array(array(Database::getConnection($this->__master), $method), $args);
  }

  public function __get($key) {
    return Database::getConnection($this->__master)->$key;
  }

  public function __set($key, $value) {
    Database::getConnection($this->__master)->$key = $value;
  }






  /**
   * Internal autoslave functions
   */

  function getMasterConnection() {
    return Database::getConnection($this->__master);
  }

  static function isTargetAvailable($target, $key = NULL) {
    $conn = Database::getConnection($target, $key);
    return $conn->getTarget() == $target;
  }

  function prepareAutoslaveTarget(&$options) {
    if (empty($options['target'])) { 
      $options['target'] = $this->getTarget();
      return FALSE;
    }
    if ($options['target'] == 'slave') {
      // Direct to the autoslave if 'slave' is explicitely chosen.
      $options['target'] = $this->__slave;
    }
    if (!self::isTargetAvailable($options['target'])) {
      $options['target'] = $this->getTarget();
      return FALSE;
    }
    if ($options['target'] == $this->getTarget()) {
      return FALSE;
    }
    return TRUE;
  }

  /**
   * Determine the which target to use based on a query.
   */
  private function deriveTargetFromQuery($query) {
    $this->ensureAffectedTables();

    $is_write_query = preg_match('/^\s*('.
      'UPDATE|INSERT|REPLACE|DELETE|'.
      'ALTER|CREATE|DROP|TRUNCATE|RENAME|'.
      'BEGIN|START\s+TRANSACTION|COMMIT|ROLLBACK|'.
      'RELEASE|SAVEPOINT|'.
      '(.*FOR UPDATE$)|(.*LOCK IN SHARE MODE$)'.
    ')\b/i', $query);

    // Find all tables used in the query.
    preg_match_all('@\{(\w+)\}@', $query, $matches);
    $tables = $matches[1];

    if ($is_write_query) {
      // Even if forceMaster() is true, we still need to tag tables that have been written to,
      // in case we go back to forceMaster() false later on.
      $this->addAffectedTables($tables);
      return $this->__master;
    }
    elseif ($this->forceMaster()) {
      return $this->__master;
    }
    else {
      return array_intersect($tables, $this->__tables) ? $this->__master : $this->__slave;
    }

  }

  /**
   * Get the target for a table
   *
   * @param $table
   *   Table to get target for
   * @return string
   *   Target
   */
  private function getTargetForTable($table) {
    $this->ensureAffectedTables();
    return $this->forceMaster() || isset($this->__tables[$table]) ? $this->__master : $this->__slave;      
  }

  /**
   * Get/set force master counter
   */
  public function forceMaster($force_master = NULL) {
    if (isset($force_master)) {
      $this->__force_master += $force_master;
    }
    return $this->__force_master;
  }

  /**
   * Redeclare the master target as the default target.
   */
  public function hardSwitch() {
    $connection_info = Database::getConnectionInfo('default');
    Database::renameConnection('default', 'autoslave_original_default');
    Database::addConnectionInfo('default', 'default', $connection_info[$this->__master]);
  }

  /**
   * Get list of affected tables.
   */
  public function getAffectedTables() {
    return $this->__tables;
  }

  /**
   * Add tables to list of affected tables.
   */
  public function addAffectedTables($tables, $update_session = TRUE) {
    $this->__tables = array_unique(array_merge($this->__tables, $tables));
    if ($update_session) {
      foreach ($tables as $table) {
        // Reflag tables with timestamp later, if we're inside a transaction.
        $_SESSION['autoslave_affected_tables'][$this->getTarget()][$table] = $this->transactionDepth() ? 0 : time() + variable_get('autoslave_assumed_replication_lag', AUTOSLAVE_ASSUMED_REPLICATION_LAG);
      }
    }
    return $this->__tables;
  }

  public function addAffectedTable($table, $update_session = TRUE) {
    return $this->addAffectedTables(array($table => $table), $update_session);
  }

  /**
   * Ensure that tables affected by write from previous requests are flagged,
   * so that queries for these tables will go to the master.
   */
  function ensureAffectedTables() {
    static $has_run = FALSE;
    if ($has_run) {
      return;
    }

    if (!function_exists('drupal_session_started') || !drupal_session_started()) {
      return;
    }

    $has_run = TRUE;

    // Use connection to master for affected tables within the given assumed lag time interval.
    if (isset($_SESSION['autoslave_affected_tables'])) {
      // We use server request time instead of time() for the sake of db's with isolation level snapshots.
      foreach ($_SESSION['autoslave_affected_tables'] as $target => $tables) {
        $connection = Database::getConnection($target);
        if ($connection->driver() != 'autoslave') {
          continue;
        }

        foreach ($tables as $table => $expire) {
          if ($_SERVER['REQUEST_TIME'] <= $expire) {
            $connection->addAffectedTable($table, FALSE);
          }
          else {
            unset($_SESSION['autoslave_affected_tables'][$target][$table]);
          }
        }
        // If no affected tables, remove the variable from session.
        if (empty($_SESSION['autoslave_affected_tables'][$target])) {
          unset($_SESSION['autoslave_affected_tables'][$target]);
        }
      }

      // If no affected tables, remove the variable from session.
      if (empty($_SESSION['autoslave_affected_tables'])) {
        unset($_SESSION['autoslave_affected_tables']);
      }
    }
  }
}

/**
 * Handle force master for transactions
 */
class DatabaseTransaction_autoslave extends DatabaseTransaction {
  public $autoslave_connection;
  protected $parent;

  public function __construct(DatabaseConnection $connection, $name = NULL) {
    $this->autoslave_connection = $connection;
    $this->autoslave_connection->forceMaster(1);
    $this->parent = $this->autoslave_connection->getMasterConnection()->startTransaction($name);
  }

  public function __destruct() {
    unset($this->parent);
    $this->autoslave_connection->forceMaster(-1);
    // If tables were flagged inside a transaction, update their timestamp to the commit time.
    if ($this->autoslave_connection->transactionDepth() == 0) {
      foreach ($_SESSION['autoslave_affected_tables'][$this->autoslave_connection->getTarget()] as $table => &$expire) {
        $expire = $expire ? $expire : time() + variable_get('autoslave_assumed_replication_lag', AUTOSLAVE_ASSUMED_REPLICATION_LAG);
      }
    }
  }

  public function name() { return $this->parent->name(); }
  public function rollback() { return $this->parent->rollback(); }
}

